/*
	Countdown Version 3.1 - The Updated Return
	Based on the popular MKWii Countdown mod by MrBean35000VR
	Created by Latius
	Thanks to Afrorule, Kitoko, Aquatic, Lat', and a few others who had helped with this mod at one point or another
	
*/

freeslot("sfx_cntcr", "sfx_cntlo", "sfx_cntko", "sfx_cntbt", "sfx_cntdv")

local version = "V3.1"
rawset(_G, "FLASHTIME", 5 * TICRATE / 2)
local MAXLAPS = 45 //Max Laps. Can NOT be more than 48.

local counton = CV_RegisterVar({"cd_enabled", "On", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})

// Starting time limit in seconds.
local counttime = CV_RegisterVar({"cd_time", "120", CV_NETVAR, {MIN = 60, MAX = 300}})

// Attack Time Bonus variables.
local countattack = CV_RegisterVar({"cd_attack", "On", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local countattackcap = CV_RegisterVar({"cd_attackcap", "On", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local countattackbonus = CV_RegisterVar({"cd_attackbonus", "6", CV_NETVAR, {MIN = 1, MAX = 15}})
local countattackmax = CV_RegisterVar({"cd_attackmax", "10", CV_NETVAR, {MIN = 1, MAX = 30}})

// Lap/Section Time Bonus variables.
local countlap = CV_RegisterVar({"cd_lap", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local countlapbonus = CV_RegisterVar({"cd_lapbonus", "15", CV_NETVAR, {MIN = 5, MAX = 30}})
local COUNTLAPMIN = 2
local countlapdecay = CV_RegisterVar({"cd_lapdecay", "2", CV_NETVAR, {MIN = 1, MAX = 5}})

// Drift Time Bonus variables.
local countdrift = CV_RegisterVar({"cd_drift", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local countdriftsolo = CV_RegisterVar({"cd_driftsolo", "On", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})

// Perfect Start Time Bonus variables
local countstart = CV_RegisterVar({"cd_start", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})

local colouredhud = CV_RegisterVar({"cd_coloredhud", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local countmeme = CV_RegisterVar({"cd_meme", "Off", CV_SHOWMODIF, CV_OnOff})

local numplayers
local outplayers
local endtimer

//3.0 Update: all the variables are now here. Plus there's some override flags,
//so if another mod changes them, they can set the flags to lock it to the
//values set by the mod instead of through this.
rawset(_G, "countdownmode", {
	version = "V3.1",
	enabled = true, //Is Countdown even on?
	base = { //The Basics
		overrride = false,
		colouredhud = false,
		time = 90, //The Base Time Limit in Seconds.
		lapcap = 45 //Replacing MAXLAPS
	},
	attack = { //Attack or Protect
		overrride = false,
		enabled = true,
		capped = true, //Do we have a cap to how many times someone can gain this bonus?
		timeplus = 5, //What's the bonus?
		cap = 10 //The max number of attacks before this bonus is capped.
	},
	laps = { //Complete Laps
		overrride = false,
		enabled = false,
		timeplus = 15, //The basic time gained per lap completed.
		decay = 2, //How many laps until the time decays by 1s?
		mintime = COUNTLAPMIN //The lowest the time bonus per lap can get.
	},
	drift = { //DEJA VU!
		overrride = false,
		enabled = false,
		soloenabled = true, //Because Solo Mode doesn't allow attacking, should they get drift bonuses instead?
		timeplus = {TICRATE, TICRATE*3, TICRATE*5} //Blue, Red and Rainbow Time Bonus (in frames)
	},
	start = { //Boost To Win!
		overrride = false,
		enabled = false,
		timeplus = {TICRATE*2, TICRATE*5} //Good and Perfect Start Time Bonus (in frames)
	}
})

local function resetcountdown()
	if counton.value then countdownmode.enabled = true else countdownmode.enabled = false end
	if not countdownmode.base.override then
		countdownmode.base.time = counttime.value
		countdownmode.base.colouredhud = (colouredhud.value > 0)
		countdownmode.base.lapcap = MAXLAPS
	else
		countdownmode.base.time = min(300, max(60, $)) //Must be between 1 and 5 minutes
		countdownmode.base.lapcap = min(48, max(10, $)) //Must be between 10 and 48 laps as the cap
	end
	if not countdownmode.attack.override then
		countdownmode.attack.enabled = (countattack.value > 0)
		countdownmode.attack.capped = (countattackcap.value > 0)
		countdownmode.attack.timeplus = countattackbonus.value
		countdownmode.attack.cap = countattackmax.value
	else
		countdownmode.attack.timeplus = min(15, max($, 1))
		countdownmode.attack.cap = min(30, max($, 1))
	end
	if not countdownmode.laps.override then
		countdownmode.laps.enabled = (countlap.value > 0)
		countdownmode.laps.timeplus = countlapbonus.value
		countdownmode.laps.decay = countlapdecay.value
		countdownmode.laps.mintime = COUNTLAPMIN
	else
		countdownmode.laps.timeplus = min(30, max(5, $))
		countdownmode.laps.decay = min(5, max($, 1))
		countdownmode.laps.mintime = min(max(COUNTLAPMIN, $), countdownmode.laps.timeplus - 2) 
		//mintime must be at least 2s, and at most 2s less than the starting time
	end
	if not countdownmode.drift.override then
		countdownmode.drift.enabled = (countdrift.value > 0)
		countdownmode.drift.soloenabled = (countdriftsolo.value > 0)
		countdownmode.drift.timeplus = {TICRATE, TICRATE*3, TICRATE*5}
	else
		for i = 1, 3 do
			countdownmode.drift.timeplus[i] = TICRATE*min(5*i,max($/TICRATE, i))
		end
	end
	if not countdownmode.start.override then
		countdownmode.start.enabled = (countstart.value > 0)
		countdownmode.start.timeplus = {TICRATE*2, TICRATE*5}
	else
		for i = 1, 2 do
			countdownmode.start.timeplus[i] = TICRATE*min(5*i,max($/TICRATE, i))
		end
	end
end

COM_AddCommand("CountWelcome", function(p)
	CONS_Printf(p, "Welcome to Countdown " .. countdownmode.version .. " which alters")
	CONS_Printf(p, "the normal race mode into a race against the clock.")
	CONS_Printf(p, "")
	CONS_Printf(p, "For help, please use the command CountHelp in the console.")
end)

COM_AddCommand("CountHelp", function(p)
	CONS_Printf(p, "NetVars:")
	CONS_Printf(p, "cd_time - the starting time of a race in seconds. Default 90, Range 60-300")
	CONS_Printf(p, "cd_attack - determines if attacks adds time. Default On.")
	CONS_Printf(p, "cd_attackcap - determines if there's a limit to attack time bonuses. Default On.")
	CONS_Printf(p, "cd_attackbonus - the number of seconds gained from an attack. Default 5, Range 1-15.")
	CONS_Printf(p, "cd_attackmax - determines the cap of how many attacks can score time. Default 10, Range 1-30.")
	CONS_Printf(p, "cd_lap - determines if there's bonus time for completing a lap or section. Default Off.")
	CONS_Printf(p, "cd_lapbonus - determines the base time bonus for completing a lap or sector. Default 15, Range 5-30.")
	CONS_Printf(p, "cd_lapdecay - determines the number of laps before the bonus decays by 1 second. Default 2, Range 1-5.")
	CONS_Printf(p, "cd_drift - determines if there's bonus time for using drifts. Default Off.")
	CONS_Printf(p, "cd_start - determines if there's a bonus for getting a starting boost. Default On.")
	CONS_Printf(p, "cd_colouredhud - determines if the HUD is coloured according to the skin or team colour. Default Off.")
	CONS_Printf(p, "Commands:")
	CONS_Printf(p, "CountSettings - shows the current settings")
end)

COM_AddCommand("CountSettings", function(p)
	CONS_Printf(p, "Current Settings:")
	CONS_Printf(p, "Base Starting Time: " .. (countdownmode.base.time) / 60 .. "min " .. (countdownmode.base.time) % 60 .. "s")
	if countdownmode.attack.enabled then
		CONS_Printf(p, "Time bonuses from attacks ON")
		CONS_Printf(p, "Bonus: +" ..  countdownmode.attack.timeplus    .. "s per attack")
		CONS_Printf(p, "       +" .. (countdownmode.attack.timeplus*3) .. "s for Sink attacks")
		if countdownmode.attack.capped then
			CONS_Printf(p, "Attack Bonus cap is ON")
			CONS_Printf(p, "Cap: " .. countdownmode.attack.cap .. " attacks")
		else
			CONS_Printf(p, "Attack Bonus cap is OFF")
		end
	else 
		CONS_Printf(p, "Time bonuses from attacks DISABLED")
	end
	if countdownmode.laps.enabled then
		CONS_Printf(p, "Time bonuses from laps ON")
		CONS_Printf(p, "Base Lap Bonus: " .. countdownmode.laps.timeplus .. "s")
		CONS_Printf(p, "Lap Bonuses will decay 1s after every " .. countdownmode.laps.decay .. " laps")
	else
		CONS_Printf(p, "Time bonuses from laps OFF")
	end
	if countdownmode.drift.enabled then
		CONS_Printf(p, "Time bonuses from drifts ON")
		if countdownmode.drift.override then
			CONS_Printf(p, "Blue: " .. countdownmode.drift.timeplus[1]/TICRATE .. "s bonus, Red: " .. countdownmode.drift.timeplus[2]/TICRATE .. "s bonus, Rainbow: " .. countdownmode.drift.timeplus[3]/TICRATE .. "s bonus")
		else
			CONS_Printf(p, "Blue: 1s bonus, Red: 3s bonus, Rainbow: 5s bonus")
		end
	else
		CONS_Printf(p, "Time bonuses from drifts OFF")
	end
	if countdownmode.start.enabled then
		CONS_Printf(p, "Time bonuses from start boosts ON")
		if countdownmode.start.override
			CONS_Printf(p, "Start Boost: ".. countdownmode.start.timeplus[1] .."s bonus")
			CONS_Printf(p, "Perfect Start Boost: ".. countdownmode.start.timeplus[2] .."s bonus")
		else
			CONS_Printf(p, "Start Boost: 2s bonus")
			CONS_Printf(p, "Perfect Start Boost: 5s bonus")
		end
	else
		CONS_Printf(p, "Time bonuses from start boosts OFF")
	end
end)

local function padTimerZero(num)
	if num < 10
		return "0" .. tostring(num)
	else
		return tostring(num)
	end
end

local function ticsToTimer(num, bonus)
	if bonus then
		return "+" .. G_TicsToSeconds(num) .. "\""
	else
		return padTimerZero(G_TicsToMinutes(num, true)) .. "'" ..  padTimerZero(G_TicsToSeconds(num)) .. "\"" .. padTimerZero(G_TicsToCentiseconds(num))
	end
end

local function drawLetter(v, font, char, x, y, f, c)
	local asc = string.byte(char)
	local padding = "000"
	if asc > 99
		padding = ""
	elseif asc > 9
		padding = "0"
	elseif asc > 0
		padding = "00"
	end
	v.draw(x, y, v.cachePatch(font .. padding .. tostring(asc)), f, c)
end

local function drawFontString(v, font, str, x, y, spacing, f, c)
	if str == nil
	or string.len(str) == 0
		str = "NIL"
	end
	for a=1,string.len(str) do
		drawLetter(v, font, string.sub(str, a, a), x + ((a - 1) * spacing), y, f, c)
	end
end

local function startset(player)
	player.countdown = {}
	player.countdown.time = countdownmode.base.time*TICRATE
	player.countdown.attacks = 0
	player.countdown.laps = 0
	player.countdown.place = 0 // Used to store the finish placement
	player.countdown.lapbonus = 0
	player.countdown.attackbonus = 0
	player.countdown.sinkhit = false //Did they hit someone with the Sink?
	player.countdown.driftbonus = 0
	player.countdown.drifttype = 0
	player.countdown.lastdriftboost = 0
	player.countdown.startboost = 0
	player.countdown.startbonus = 0
	player.countdown.out = false
	player.countdown.haddrift = 0
	player.countdown.trick = 0
	player.countdown.trickbonus = 0
end

addHook("PlayerExplode", function(player, inflictor, source)
	if (not (gametype & GT_MATCH)) and countdownmode.enabled and countdownmode.attack.enabled then
		//print("Player: "..player.name)
		//if source and source.player then print("Source: "..source.player.name) end
		//if inflictor and inflictor.player then print("Inflictor: "..source.player.name) end
		if source and (source ~= player.mo) and source.player and (source.player.countdown.time >= source.player.realtime) then
			if not (countdownmode.attack.capped and (countdownmode.attack.cap <= source.player.countdown.attacks)) then
				source.player.countdown.attackbonus = FLASHTIME
				source.player.countdown.time = $ + (TICRATE * countdownmode.attack.timeplus)
				source.player.countdown.sinkhit = false
				if countdownmode.attack.capped then
					source.player.countdown.attacks = $ + 1
				end
			end
		end
	end
end)

addHook("PlayerSpin", function(player, inflictor, source)
	if (not (gametype & GT_MATCH)) and countdownmode.enabled and countdownmode.attack.enabled then
		//print("Player: "..player.name)
		//if source and source.player then print("Source: "..source.player.name) end
		//if inflictor and inflictor.player then print("Inflictor: "..source.player.name) end
		if source and (source ~= player.mo) and source.player and (source.player.countdown.time >= source.player.realtime) then
			if not (countdownmode.attack.capped and (countdownmode.attack.cap <= source.player.countdown.attacks)) then
				source.player.countdown.attackbonus = FLASHTIME
				source.player.countdown.time = $ + (TICRATE * countdownmode.attack.timeplus)
				source.player.countdown.sinkhit = false
				if countdownmode.attack.capped then
					source.player.countdown.attacks = $ + 1
				end
			end
		end
	end
end)

addHook("PlayerSquish", function(player, inflictor, source)
	if (not (gametype & GT_MATCH)) and countdownmode.enabled and countdownmode.attack.enabled then
		//print("Player: "..player.name)
		//if source and source.player then print("Source: "..source.player.name) end
		//if inflictor and inflictor.player then print("Inflictor: "..source.player.name) end
		if source and (source ~= player.mo) and source.player and (source.player.countdown.time >= source.player.realtime) then
			if not (countdownmode.attack.capped and (countdownmode.attack.cap <= source.player.countdown.attacks)) then
				source.player.countdown.attackbonus = FLASHTIME
				source.player.countdown.time = $ + (TICRATE * countdownmode.attack.timeplus)
				source.player.countdown.sinkhit = false
				if countdownmode.attack.capped then
					source.player.countdown.attacks = $ + 1
				end
			end
		end
	end
end)

//Sink Hits are special! They don't count towards the cap, and they gain you even more time!
addHook("MobjDamage", function(player, inflictor, source)
	if (not (gametype & GT_MATCH)) and countdownmode.enabled and countdownmode.attack.enabled then
		if (inflictor and inflictor.type == MT_SINK)
			if source and (source ~= player.mo) and source.player and (source.player.countdown.time >= source.player.realtime) then
				source.player.countdown.attackbonus = FLASHTIME
				source.player.countdown.time = $ + (TICRATE * countdownmode.attack.timeplus * 3)
				source.player.countdown.sinkhit = true
			end
		end
	end
end)

addHook("NetVars", function(network)
	endtimer = network(endtimer)
	numplayers = network(numplayers)
	outplayers = network(outplayers)
	countdownmode = network(countdownmode)
end)

addHook("MapChange", do
	endtimer = 0
	outplayers = 0
	resetcountdown()
end)

addHook("MapLoad", do
	resetcountdown()
	if not (gametype & GT_MATCH) then
		local lapmax 
		if mapheaderinfo[gamemap].levelflags & 32 then
			lapmax = mapheaderinfo[gamemap].numlaps
		else
			lapmax = countdownmode.base.lapcap
		end
		if countdownmode.enabled then
			COM_BufInsertText(server, "numlaps " .. lapmax + 2)
			COM_BufInsertText(server, "basenumlaps " .. lapmax + 2)
			for p in players.iterate do
				if p.mo and p.mo.valid
					startset(p)
				end
			end	
		else
			COM_BufInsertText(server, "numlaps " .. mapheaderinfo[gamemap].numlaps)
			COM_BufInsertText(server, 'basenumlaps "Map default"')
		end
		outplayers = 0
		endtimer = 0
	end
end)

//Moving these bonus stuff to here so that other mods could call them
rawset(_G, "cd_startbonus", function(p)
	if not (p and p.valid and p.countdown) then return end
	if p.kartstuff[k_sneakertimer] > 68 then
		p.countdown.startboost = 2
		p.countdown.startbonus = FLASHTIME
	elseif p.kartstuff[k_startboost] > 0 then
		p.countdown.startboost = 1
		p.countdown.startbonus = FLASHTIME
	end
	if p.countdown.startboost > 0 then
		p.countdown.time = $ + countdownmode.start.timeplus[p.countdown.startboost]
	end
end)

rawset(_G, "cd_lapbonus", function(p)
	if not (countdownmode.enabled and p and p.valid and p.countdown and (p.countdown.time >= p.realtime)) then return end
	local timebonus
	timebonus = countdownmode.laps.timeplus - (p.countdown.laps / countdownmode.laps.decay)
	if timebonus < countdownmode.laps.mintime then timebonus = countdownmode.laps.mintime end
	p.countdown.time = $ + (timebonus * TICRATE)
	p.countdown.lapbonus = FLASHTIME
	S_StartSound(p.mo,sfx_cntbt,p)
end)

rawset(_G, "cd_driftbonus", function(p)
	if not (countdownmode.enabled and p and p.valid and p.countdown and (p.countdown.time >= p.realtime)) then return end
	p.countdown.lastdriftboost = p.kartstuff[k_driftboost]
	if p.countdown.lastdriftboost > 50 then
		p.countdown.drifttype = 3
		if countmeme.value == 1 then
			S_StartSound(p.mo,sfx_cntdv,p)
		else
			S_StartSound(p.mo,sfx_cntbt,p)
		end
	elseif p.countdown.lastdriftboost > 20 then
		p.countdown.drifttype = 2
		S_StartSound(p.mo,sfx_cntbt,p)
	else
		p.countdown.drifttype = 1
		//S_StartSound(p.mo,sfx_cntbt,p)
	end
	p.countdown.time = $ + countdownmode.drift.timeplus[p.countdown.drifttype]
	p.countdown.driftbonus = FLASHTIME
	p.countdown.haddrift = $ + 1
end)

//Ready to lock onto Acrobatics
rawset(_G, "cd_trickbonus", function(p, tier, bonustable)
	if not (countdownmode.enabled and p and p.valid and p.countdown and (p.countdown.time >= p.realtime)) then return end
	local tricktier
	if tier ~= nil then tier = tricktier
	elseif p.trickgrade then tier = p.trickgrade //So that Acrobatics doesn't need to supply this
	else return end //How am I supossed to know how to get the bonus without the tier?!?
	
	local table
	if bonustable then table = bonustable
	elseif p.trickgrade then table = { //So that Acrobatics doesn't need to supply this
		[-1] = 0,
		[0]  = 0,
		[1]  = TICRATE,
		[2]  = TICRATE,
		[3]  = TICRATE*2,
		[4]  = TICRATE*2,
		[5]  = TICRATE*3
	}
	else return end //How am I supossed to know how to get the bonus without a table?!?
	
	if table[tier] == nil then return end //I can't compare it if it doesn't exist!
	
	if table[tier] > 0 then //We have a bonus, so use it.
		local secbonus 
		if table[tier] >= TICRATE then secbonus = max(table[tier]/TICRATE, 1)
		else secbonus = table[tier] end
		if secbonus > 3*tier then secbonus = 3*tier end //Don't get too greedy!
		p.countdown.time = $ + (secbonus*TICRATE)
		p.countdown.trick = secbonus
		p.countdown.trickbonus = FLASHTIME
	end
end)

addHook("ThinkFrame", do
	if not countdownmode.enabled then return end
	if not (gametype & GT_MATCH) then
		local soloplayer
		if (numplayers == nil) then numplayers = 0 end
		if numplayers <= 1 then soloplayer = true else soloplayer = false end
		numplayers = 0
		outplayers = 0
		
		if endtimer == nil then endtimer = 0 end
		for p in players.iterate do
			if p.spectator then continue end //Skip spectators
			numplayers = $ + 1
			
			p.kartstuff[k_lapanimation] = 0 //Thanks Lat' for showing me that this is possible.
			
			if (p.countdown == nil) or ((p.countdown.place > 0) and (not p.exiting)) then
				//Reset everything
				startset(p)
				if endtimer ~= 0 then
					endtimer = 0
				end
			end
			
			if not (p.countdown.time >= p.realtime) then //If the player is out of time
				//Make sure they can't move, but can still be hit by items
				p.pflags = $ | PF_STASIS
				outplayers = $ + 1
				if p.countdown.time == p.realtime and not p.countdown.out then
					//Play a SFX to give an audio cue that time's up
					S_StartSound(p.mo,sfx_cntko,p)
					p.countdown.out = true
				end
				continue //Move to the next player
			end
			
			if p.exiting
				or ((p.countdown.laps == mapheaderinfo[gamemap].numlaps) and (mapheaderinfo[gamemap].levelflags & 32))
				or p.countdown.laps == countdownmode.base.lapcap then //Completed all the laps, so they are 'out'
				if not p.exiting then 
					p.exiting = 1
					S_StartSound(p.mo,sfx_lvpass,p)
					if p.kartstuff[k_position] == 1 //Winner
						S_ChangeMusic("KRWIN",true,p)
						S_StartSound(p.mo,sfx_kwin,p)
					elseif K_IsPlayerLosing(p) //Loser
						S_ChangeMusic("KRLOSE",true,p)
						S_StartSound(p.mo,sfx_klose,p)
					else //OK
						S_ChangeMusic("KROK",true,p)
						S_StartSound(p.mo,sfx_kwin,p)
					end
					p.countdown.place = p.kartstuff[k_position]
				end
				outplayers = $ + 1
				continue //Move to the next player
			end
			
			//Perfect Start Bonus!
			if (p.realtime == 1) and countdownmode.start.enabled then
				cd_startbonus(p)
			elseif p.countdown.startbonus > 0 then
				p.countdown.startbonus = $ - 1
			end
			
			//Lap Bonus!
			if (p.laps > p.countdown.laps) then
				p.countdown.laps = $ + 1
				if countdownmode.laps.enabled
				and not (((p.countdown.laps == mapheaderinfo[gamemap].numlaps) and (mapheaderinfo[gamemap].levelflags & 32)) or p.countdown.laps == countdownmode.base.lapcap) then
					cd_lapbonus(p)
				end
			elseif (p.countdown.lapbonus > 0) then //Reduce countdown on Lap Bonus
				p.countdown.lapbonus = $ - 1
			end
			
			//Drift Bonus!
			if (p.kartstuff[k_driftboost] > p.countdown.lastdriftboost)
			and (countdownmode.drift.enabled or (soloplayer and countdownmode.drift.soloenabled))
			and p.countdown.haddrift < 3 then
				cd_driftbonus(p)
			elseif p.countdown.driftbonus > 0 then
				p.countdown.driftbonus = $ - 1
				if p.countdown.lastdriftboost > 0 then p.countdown.lastdriftboost = $ - 1 end
				if p.countdown.haddrift then p.countdown.haddrift = 0 end
			elseif p.countdown.lastdriftboost > 0 then
				p.countdown.lastdriftboost = $ - 1
				if p.countdown.haddrift then p.countdown.haddrift = 0 end
			end
			
			if p.countdown.attackbonus > 0 then
				p.countdown.attackbonus = $ - 1
			end
			
			if p.countdown.trickbonus > 0 then
				p.countdown.trickbonus = $ - 1
			end
			
			if (p.countdown.time - (10 * TICRATE)) == p.realtime then
				//Play a warning SFX to give an audio cue that time is about to run out.
				S_StartSound(p.mo,sfx_cntcr,p)
			elseif (p.countdown.time - (10 * TICRATE)) < p.realtime
				and (p.realtime % (TICRATE)) == 0
				and not (p.countdown.time == p.realtime) then
				//Time's running out!
				S_StartSound(p.mo,sfx_cntlo,p)
			end
		end
	
		if endtimer > 0 then
			endtimer = $ + 1
			if endtimer > (5 * TICRATE) then
				G_ExitLevel() //After 5 seconds of the end stuff, end the level
			end
		end
	
		if (numplayers == outplayers) and (numplayers > 0) and (endtimer == 0) then //Race over!
			for p in players.iterate do
				if not p.exiting then //Have they not finished the laps/race?
					if p.kartstuff[k_position] == 1 //Winner
						S_ChangeMusic("KRWIN",true,p)
						S_StartSound(p.mo,sfx_kwin,p)
					elseif K_IsPlayerLosing(p) //Loser
						S_ChangeMusic("KRLOSE",true,p)
						S_StartSound(p.mo,sfx_klose,p)
					else //OK
						S_ChangeMusic("KROK",true,p)
						S_StartSound(p.mo,sfx_kwin,p)
					end
					p.countdown.place = p.kartstuff[k_position]
					p.realtime = $ + (p.kartstuff[k_position] - 1) //Set like this until custom end screen stats can be made
					p.exiting = 1
				end
			end
			endtimer = 1
		end
	end
end)

hud.add(function(v, p)
	if (not (gametype & GT_MATCH)) and countdownmode.enabled then
		hud.disable("time") //Timer is made custom
		hud.disable("gametypeinfo") //Same with the lap counter
		hud.disable("speedometer") //Don't want this for now
		local xpos = 9
		local ypos = 170
		local sypos = 90
		local fontstring = "MKFNT"
		local kofontstring = "STCFN"
		if countdownmode.base.colouredhud == 1 then
			kofontstring = "SCGFT"
			if splitscreen > 1 then
				fontstring = "SCGFT"
			else
				fontstring = "MKGFT"
			end
		elseif splitscreen > 1 then
			fontstring = "STCFN"
		end
		
		local splitx
		local splity
		if splitscreen then
			local n = c.pnum
			if not n return end	
		
			if splitscreen == 1 then
				splity = n
				splitx = nil
			else
				if n <= 2 then splity = 1 else splity = 2 end
				if n == 1 or n == 3 then splitx = 1 else splitx = 2 end
			end
		else
			splity = nil
			splitx = nil
		end
		
		local colormap
		local lapcount
		if not p.exiting and (p.countdown.time > p.realtime) then
			lapcount = tostring(p.countdown.laps)
			if p.mo and p.mo.valid then
				colormap = v.getColormap(0, p.mo.color)
			else
				colormap = v.getColormap(0, p.color)
			end
		elseif (p.countdown.laps == countdownmode.base.lapcap) or ((p.countdown.laps == mapheaderinfo[gamemap].numlaps) and (mapheaderinfo[gamemap].levelflags & 32)) then
			lapcount = "FIN"
			colormap = v.getColormap(0, p.color)
		else
			lapcount = "OUT"
			colormap = v.getColormap(0, p.color)
		end
		
		local attackstring = ""
		if countdownmode.attack.enabled then
			attackstring = p.countdown.attacks
			if countdownmode.attack.capped then
				attackstring = $.."/"..countdownmode.attack.cap.." KOS"
				if countdownmode.attack.cap == p.countdown.attacks then attackstring = "MAX KOS" end
			else
				attackstring = $.." KOS"
			end
		end
		
		if splitscreen < 2 then
			if splitscreen then ypos = $ - ((1 - splity) * 100) end
			v.draw(xpos, ypos, v.cachePatch("K_STLAPS"), V_HUDTRANS)
			xpos = $ + 36
			ypos = $ + 3
			if p.spectator then //Skip spectators
				drawFontString(v, "MKFNT", "N/A", xpos, ypos, 12, V_HUDTRANS)
			else
				drawFontString(v, fontstring, lapcount, xpos, ypos, 12, V_HUDTRANS, colormap)
				if attackstring ~= "" then
					drawFontString(v, kofontstring, attackstring, xpos-32, ypos + 16, 8, V_HUDTRANS, colormap)
				end
			end
		elseif not p.spectator then
			xpos = $ - 5 + (160 * (splitx - 1))
			sypos = $ + (100 * (splity - 1))
			if attackstring ~= "" then
				sypos = $ - 10
				drawFontString(v, kofontstring, attackstring, xpos, sypos+10, 8, V_HUDTRANS, colormap)
			end
			drawFontString(v, fontstring, "LAP:"..lapcount, xpos, sypos, 8, V_HUDTRANS, colormap)
		end
		
		local timex = 202
		local stimex = 88
		local timey = 0
		local stimey = 0
		if not p.spectator then
			local displaytime = p.countdown.time - p.realtime
			if displaytime < 0 then displaytime = 0 end
			if not splitscreen then
				v.draw(timex, timey, v.cachePatch("K_STTIME"), V_HUDTRANS)
				if ((displaytime > 10 * TICRATE) or (p.realtime % 10 > 4) or (displaytime == 0)) then
					drawFontString(v, fontstring, ticsToTimer(displaytime, false), timex + 10, timey + 3, 12, V_HUDTRANS, colormap)
				else
					drawFontString(v, "MKGFT", ticsToTimer(displaytime, false), timex + 10, timey + 3, 12, V_HUDTRANS, v.getColormap(0, SKINCOLOR_RED), true)
				end
			elseif splitscreen == 1 then
				v.draw(timex, stimey + ((1 - splity) * 100), v.cachePatch("K_STTIME"), V_HUDTRANS)
				if ((displaytime > 10 * TICRATE) or (p.realtime % 10 > 4) or (displaytime == 0)) then
					drawFontString(v, fontstring, ticsToTimer(displaytime, false), timex + 10, stimey + ((splity - 1) * 100) + 3, 12, V_HUDTRANS, colormap)
				else
					drawFontString(v, "MKGFT", ticsToTimer(displaytime, false), timex + 10, stimey + ((splity - 1) * 100) + 3, 12, V_HUDTRANS, v.getColormap(0, SKINCOLOR_RED), true)
				end
			elseif ((displaytime > 10 * TICRATE) or (p.realtime % 10 > 4) or (displaytime == 0)) //splitscreen == 2 or 3
				drawFontString(v, fontstring, ticsToTimer(displaytime, false), stimex + (160 * (splitx - 1)), stimey + (100 * (splity - 1)) + 3, 8, V_HUDTRANS, colormap)
			else
				drawFontString(v, "SCGFT", ticsToTimer(displaytime, false), stimex + (160 * (splitx - 1)), stimey + (100 * (splity - 1)) + 3, 8, V_HUDTRANS, v.getColormap(0, SKINCOLOR_RED), true)
			end
		else
			if not splitscreen then
				v.draw(timex, timey, v.cachePatch("K_STTIME"), V_HUDTRANS)
				if (p.realtime % 35 < 18) then
					drawFontString(v, "MKFNT", "SPECTATE", 212, 3, 12, V_HUDTRANS, colormap)
				end
			elseif splitscreen == 1 then
				v.draw(timex, stimey + ((splity - 1) * 100), v.cachePatch("K_STTIME"), V_HUDTRANS)
				if (p.realtime % 35 < 18) then
					drawFontString(v, "MKFNT", "SPECTATE", timex + 10, stimey + (#p * 100) + 3, 12, V_HUDTRANS, colormap)
				end
			elseif (p.realtime % 35 < 18) then //splitscreen == 2 or 3
				drawFontString(v, fontstring, "SPECTATE", stimex + (160 * (splitx - 1)), stimey + (100 * (splity - 1)) + 3, 8, V_HUDTRANS, colormap)
			end
		end
		local bonustime = 0
		local sxpos = 104
		local pos = 17
		if splitscreen == 1 then pos = $ + ((splity - 1) * 100) end
		if splitscreen > 1 then
			pos = $ - 6 + ((splity - 1) * 100)
			sxpos = $ + (160 * (splitx - 1))
		end
		if (p.countdown.startbonus > 0) and (p.realtime % 10 < 5) and (splitscreen < 2) then 
			bonustime = countdownmode.start.timeplus[p.countdown.startboost]
			drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " START", 200, pos, 12)
			pos = $ + 12
		elseif (p.countdown.startbonus > 0) and (p.realtime % 10 < 5) then
			bonustime = countdownmode.start.timeplus[p.countdown.startboost]
			drawFontString(v, fontstring, ticsToTimer(bonustime, true), sxpos + 8, pos, 8)
			pos = $ + 8
		elseif (p.countdown.lapbonus > 0) and (p.realtime % 10 < 5) and (splitscreen < 2) then 
			bonustime = max(countdownmode.laps.timeplus - ((p.laps - 1) / countdownmode.laps.decay),countdownmode.laps.mintime) * TICRATE
			if bonustime < 10*TICRATE then
				drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " LAP", 224, pos, 12)
			else
				drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " LAP", 212, pos, 12)
			end
			pos = $ + 12
		elseif (p.countdown.lapbonus > 0) and (p.realtime % 10 < 5) then
			bonustime = max(countdownmode.laps.timeplus - ((p.laps - 1) / countdownmode.laps.decay),countdownmode.laps.mintime) * TICRATE
			if bonustime >= 10*TICRATE then
				drawFontString(v, fontstring, ticsToTimer(bonustime, true), sxpos, pos, 8)
			else
				drawFontString(v, fontstring, ticsToTimer(bonustime, true), sxpos + 8, pos, 8)
			end
			pos = $ + 8
		end
		if (p.countdown.attackbonus > 0) and (p.realtime % 10 < 5) then
			bonustime = countdownmode.attack.timeplus * TICRATE
			if p.countdown.sinkhit then bonustime = $ * 3 end //Sink Hits score 3x the time!
			if (splitscreen < 2) then
				if bonustime < 10*TICRATE then
					if p.countdown.sinkhit then
						drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " SINK!", 200, pos, 12)
					else
						drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " HIT", 224, pos, 12)
					end
				else
					if p.countdown.sinkhit then
						drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " SINK!", 188, pos, 12)
					else
						drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " HIT", 212, pos, 12)
					end
				end
				pos = $ + 12
			else
				if bonustime >= 10*TICRATE then
					drawFontString(v, fontstring, ticsToTimer(bonustime, true), sxpos, pos, 8)
				else
					drawFontString(v, fontstring, ticsToTimer(bonustime, true), sxpos + 8, pos, 8)
				end
				pos = $ + 8
			end
		end
		if (p.countdown.driftbonus > 0) and (p.realtime % 10 < 5) then
			bonustime = countdownmode.drift.timeplus[p.countdown.drifttype]
			if (splitscreen < 2) then
				drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " DRIFT", 200, pos, 12)
				pos = $ + 12
			else
				drawFontString(v, fontstring, ticsToTimer(bonustime, true), sxpos + 8, pos, 8)
				pos = $ + 8
			end
		end
		if (p.countdown.trickbonus > 0) and (p.realtime % 10 < 5) then
			bonustime = p.countdown.trick * TICRATE
			if (splitscreen < 2) then
				drawFontString(v, fontstring, ticsToTimer(bonustime, true) .. " TRICK", 200, pos, 12)
				pos = $ + 12
			else
				drawFontString(v, fontstring, ticsToTimer(bonustime, true), sxpos + 8, pos, 8)
				pos = $ + 8
			end
		end
	else
		if p.numwins == nil then hud.enable("time") end
		hud.enable("gametypeinfo")
		hud.enable("speedometer")
	end
end)

hud.add(function(v)
	v.drawString(320, 192, "Countdown: "..countdownmode.version.." ", V_ALLOWLOWERCASE|V_40TRANS, "right")
end, "scores")